
plane_pose_prefix   = 'planePose_'
world_gaze_prefix   = 'planeGaze_'
coding_file         = 'coding.tsv'
target_sync_file    = 'et_sync_target_file.tsv'
VOR_sync_file       = 'VOR_sync.tsv'
validation_prefix   = 'validate_'
mapped_gaze_video   = 'mappedGaze.mp4'
gaze_export_name    = 'planeGaze'