__version__ = "1.8.0"

__title__ = "gazeMapper"
__description__ = "Automated world-based analysis of wearable eye tracker data."
__url__ = "https://github.com/dcnieho/gazeMapper"
__uri__ = __url__
__doc__ = __description__ + " <" + __url__ + ">"

__author__ = "Diederick Niehorster"
__email__ = "diederick_c.niehorster@humlab.lu.se"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2024 " + __author__
